package Data::ParseTable;

use 5.008;
use strict;
use warnings;
#use diagnostics;
use Carp;
use IO::All;
use Text::RecordParser;
use File::MMagic;
use HTML::TableExtract;
use subs qw(dbg);

use vars qw($VERSION @ISA @EXPORT @EXPORT_OK %EXPORT_TAGS);

require Exporter;

our @ISA = qw(Exporter);
our @EXPORT = qw();
our @EXPORT_OK = qw( parse_table );
%EXPORT_TAGS = ( all => \@EXPORT_OK );

our $VERSION = '0.04';

our $Debug = 0;

# Preloaded methods go here.

sub parse_table{

    my $data_in = shift;
    my $opt_in = shift;

    my $data_out_ref;
    
# Let's start by making an options hash


    my %opt = (format          => undef,  # File format (supported values are fits,rdb,html or csv)
	       field_separator => undef,  # Field sep. for 'csv' format
	       return_as_hash  => undef,  # Override default to return array of hashes
	       comment         => undef,  # Comment specifier (e.g. qr/^#/)
	       debug           => 0,
	       );

# And populating it with anything passed in

    if (defined $opt_in){
	while( my ( $option, $value ) = each %{$opt_in} ){
	    croak( __PACKAGE__, ": option error : `$option' is not recognized \n" )
		unless exists $opt{$option};

	    $opt{$option} = $value;
	}
    }
    
# Let's check that the options are relatively reasonable

    check_options(\%opt);

# if the debug option has been passed, set the package Debug = 1

    if ($opt{debug}){
	$Debug = 1;
    }

# Now let's check the to see if we've passed a filename or an array reference for
# the first argument

    my $input_type = ref($data_in);
    my $bad_data_input = __PACKAGE__ . ": parse_data not filename or array reference \n";


    $opt{array_ref_input} = ($input_type =~ /ARRAY.*/) ? 1
	                  : ($input_type =~ /\s*/ )    ? 0	
		          :                            croak($bad_data_input)
			  ;

# If passed a filename, let's confirm that the file actually exists, is a file, and 
# is readable

    if (not $opt{array_ref_input}){
#	print "Passed a Filename \n";
	croak(__PACKAGE__ , ": Not a valid or readable file!\n") 
	    unless ((-r $data_in) and (-e $data_in));
    }

# If passed an array of lines, let's confirm that they have some string data
# Actually, I'm not sure of the best way to do that.. so let's wait.
	

# If the options hash didn't tell us what kind of file it is, let's figure that out before
# moving on

    if (not defined $opt{format}) {
	dbg "Unknown Format. guessing ... ";
	$opt{format} = guess_data_format($data_in, \%opt);
    }


# Fits files are the only binaries we expect, and they should never be passed as a
# line array.  We'll use a file-reading subroutine for them.  Everybody else gets
# the works

    if ($opt{format} eq 'fits'){
	dbg "Fits file .. parsing \n";
	$data_out_ref = read_fits_table($data_in, \%opt);
    }
    else{

# if the data is already an array, pass it on, else slurp it into an array 
# and move on.

	if ($opt{array_ref_input}){
#	    print "Parsing lines of data \n";
	    $data_out_ref = read_table_data($data_in, \%opt);
	}
	else{
#	    print "Parsing a Text file \n";
	    my @data_in = make_data_array($data_in);
	    $data_out_ref = read_table_data(\@data_in, \%opt);
	}
    }

    if ($Debug){
	if ($opt{return_as_hash}){
	    table_print_hash($data_out_ref);
	}
	else{
	    table_print_array($data_out_ref);
	}
    }

    # Let's croak if our return doesn't have any columns

    # if a hash, if croak if 0 keys
    if ( $opt{return_as_hash} and ( not scalar(keys %{$data_out_ref}))){
	croak( __PACKAGE__ , ": Nothing to return \n");	
    }

    # if an array of hashes, croak if first array has 0 keys
    if ( (not $opt{return_as_hash}) and ( not scalar(keys %{$data_out_ref->[0]}))){
	croak( __PACKAGE__ , ": Nothing to return \n");
    }
    return $data_out_ref;

}


#*********************************
sub read_table_data{

# for a data table with a specified format and field separator (if needed)
# return a hash of arrays or array of hashes of that data
    
    my $data_in = shift;
    my $opt_ref = shift;

    my $format = $opt_ref->{format};
 
    my $data_out = ( $format eq 'html')     ? parse_html_table($data_in, $opt_ref)
		 : ( $format eq 'txt' )     ? parse_text_table($data_in, $opt_ref)
		 : ( $format eq 'csv' )     ? parse_text_table($data_in, $opt_ref)
		 : ( $format eq 'rdb' )     ? parse_text_table($data_in, $opt_ref)
		 :                            croak(__PACKAGE__ ,": Unknown format specified")
		 ;

    return $data_out;
}

#*********************************
sub read_fits_table{

    my $filename = shift;
    my $opt_ref = shift;

    eval 'use Astro::FITS::CFITSIO::Simple qw/ rdfits /';
    if ($@){
	croak("!$@");
    }
    
    my %fits_hash = rdfits($filename);

   
# unfortunately, that is a hash of PDLs and now I've got to figure out
# how to get them out of there.
    
    my $non_pdl_hash_ref = convert_pdl_hash(\%fits_hash);


    if ($opt_ref->{return_as_hash}){
	return $non_pdl_hash_ref;
    }
    else{
	my $fits_array_ref = table_hash_to_array($non_pdl_hash_ref);
	return $fits_array_ref;
    }
    

}


#*********************************

sub guess_data_format{

    # for an unknown file type, use the file magic to at least guess
    # the mime type and return it
    
    my $data_in = shift;
    my $opt_ref = shift;
    

    my $fm = new File::MMagic; # use internal magic file


    my $type; 
    
    my $array_in = $opt_ref->{array_ref_input};
    
    $type = $array_in      ? $fm->checktype_contents($data_in)
	  :                  $fm->checktype_filename($data_in)
	  ;

    dbg "type is $type \n";
    
    # checktype_contents seems to be missing html
    # add a check... should get rid of MMagic anyway... it isn't really
    # helping
    if ($array_in and ($type eq 'text/plain')){
	if (is_html($data_in)){
	    $type = 'text/html';
	}
	elsif(is_fits_start($data_in->[0])){
	    croak( __PACKAGE__ ,"fits files cannot be passed as arrays! " );
	}
    }

    # if it is a file and binary or plain text, check for fits file
    if ( (not $array_in) and 
	 ($type eq 'text/plain') or ($type eq 'application/octet-stream') ){
	if ( is_fits_file( $data_in ) ){
	    $type = 'fits';
	}
    }
	
    
    if ($opt_ref->{array_ref_input} and ($type eq 'application/octet-stream')){
	croak("Don't pass a binary stream\n");
    }	     

    my $format = ( $type eq 'text/html')                ? 'html'
	       : ( $type eq 'fits' )                    ? 'fits' 
	       : ( $type eq 'text/plain')               ? 'txt'
	       :                                          croak("Unknown Type\n")
	       ;
    
    
    dbg "Guessed $format! \n";
    return $format;

}

#*********************************
sub is_html{
   
    # receive a pointer to a text array and check if it matches general patterns for html
    
    my $data_ref = shift;
    my $whole_file = join(qq{},@$data_ref);    
    
    # HTMl can begin with <!DOCTYPE , <HTML, <!-- comments, or whitespace
    # I'd like to find a module that does this for me... it can also begin with
    # other funny stuff like xml
    if (($whole_file =~ /\s*<!DOCTYPE\s*HTML.*/i) or 
	($whole_file =~ /\s*<HTML>.*/i) or
	($whole_file =~ /(\s*<!--.*-->\s*)*<HTML>/i)){
	return 1;
    }
    return 0;
    
    
}

#*********************************
sub is_fits_file{

    # for a file, confirm that it is actually a fits file 
    # and define that as the format in the options hash

    my $filename = shift;
    my $opt_ref = shift;

    # only files can be fits ...
    if (defined $opt_ref->{array_ref_input}){
	return 0;
    }

    # get first 80 characters of file
    my $io = io($filename);
    my $first80 = q{};
    for my $i (0 .. 80){
	my $char = $io->getc;
	last if not defined $char;
	$first80 .= $char;
    }
	
    my $isfits = is_fits_start($first80);
    
    return $isfits;

}

#*********************************
sub is_fits_start{
    
    my $first80 = shift;

    # return 1 if matches SIMPLE
    my $isfits = ($first80 =~ /^SIMPLE\s+=\s*T.*/) ? 1 : 0 ;
    return $isfits;
}


#*********************************
sub parse_html_table{

    # for a table of html data, return a reference to a hash of that
    # data

    
    my $data_ref = shift;
    my $opt_ref = shift;

    
    my $whole_file = join(qq{},@$data_ref);

    # strip out newlines because of apparent bug in HTML::TableExtract
    $whole_file =~ s/\n//g;
    
    my $te = HTML::TableExtract->new( depth => 0 );
#    my $te = HTML::TableExtract->new();
    $te->parse($whole_file);

    # stolen code from get_iFOT_events.pl

    # Insist that exactly one table match.  (Use more robust error handling eventually)
    my @ts = $te->table_states;
#    print "tables = ", scalar(@ts), "\n";
#    croak("ERROR - more than one HTML table had the required headers") if (@ts > 1);
    croak("Could not find any HTML tables") if (@ts < 1);  # No data in table

    # Find the table coordinates and now parse the entire table, not just the 
    # spec'd header columns
    my @table_coords = $ts[0]->coords;
    my $ts = $te->table_state(@table_coords);

    # Get the data into a useful form (hash of arrays)
    my @rows = $ts->rows();
    my @cols = @{$rows[0]};

    croak("Empty table\n") if (@cols < 1);
    croak("Table contains a null column name\n") if grep {not defined} @cols;

   map { s/\A [\s<>]+ | [\s<>]+ \Z//gx } @cols;  # clean column names 
    my %data;
    foreach my $i (0 .. $#cols) {
        $data{$cols[$i]} = [ map { $_->[$i] } @rows[1..$#rows] ];
    }

    if ($opt_ref->{return_as_hash}){
	return \%data;
    }

    my $array_ref = table_hash_to_array(\%data);
    return $array_ref;

    
}



#*********************************
sub parse_text_table{

# I should address the fact that this doesn't work if passed an array without
# carriage returns

    my $data_ref = shift;
    my $opt_ref = shift;
    
    $data_ref = strip_data_comments($data_ref, $opt_ref);

    my $fs = $opt_ref->{field_separator};

    if (not defined $fs){
	$fs = guess_field_separator($data_ref);
    }
    else{
	my $n_fields = count_fields_using_fs($fs);    
	croak("Specified Field Separator does not separate file cleanly") 
	    unless ($n_fields);
    }

    # If whitespace separated, then clean leading/trailing whitespace
    if ($fs eq qr/\s+/) {
	s/^ *| *$//g for @{$data_ref};
    }

    dbg "field sep = :$fs: \n";
    $opt_ref->{field_separator} = $fs;
    
    if (($opt_ref->{format} eq 'rdb') or ($opt_ref->{field_separator} eq "\t")){
	($data_ref, $opt_ref) = check_rdb($data_ref, $opt_ref);
    }

    my $parser = Text::RecordParser->new;


    $parser->data( $data_ref );
    $parser->field_separator( $fs );
    $parser->bind_header;

    my $parsed_data = $parser->fetchall_arrayref( { Columns => {} } );

    croak("No text columns found\n") if ((keys %{$parsed_data->[0]}) < 1);
    
    if ($opt_ref->{format} ne 'rdb'){
	$parsed_data = tidy_ascii_spaces($parsed_data);
    }

    if ($opt_ref->{return_as_hash}){
	my $data_hash_ref = table_array_to_hash($parsed_data);
	return $data_hash_ref;
    }

    return $parsed_data;

}


#*********************************
sub tidy_ascii_spaces{
 
    my $in_array_ref = shift;
    my @out_array;

    foreach my $hash (@{$in_array_ref}){
	my %newhash;
	while( my ($header, $value) = each %{$hash}){
	    # strip leading or trailing whitespace off fields
	    
	    $header =~ s/^\s+//;
	    $header =~ s/\s+$//;
	    $value =~ s/^\s+//;
	    $value =~ s/\s+$//;

	    # strip external single or double quotes off a header
	    if ($header =~ /^\'(.*)\'$/){
		$header = $1;
	    }
	    elsif ($header =~ /^\"(.*)\"$/){
		$header = $1;
	    }

	    # strip external single or double quotes off a field
	    if ($value =~  /^\'(.*)\'$/){
		$value = $1;
	    }
	    elsif ($header =~ /^\"(.*)\"$/){
		$value = $1;
	    }

	    # make a new hash with the cleaned up text
	    $newhash{$header} = $value;
	}
	push( @out_array, \%newhash);
    }

    # return a reference to the newly created array
    return \@out_array;
}

#*********************************
sub make_data_array{

    my $filename = shift;
    my @data_lines = io($filename)->slurp;
    return @data_lines;

}

#*********************************
sub guess_field_separator{

    my $data_ref = shift;
#    my @field_sep = ("\t", qr/\s+/, "&", "|", ",");
    my @field_sep = (",", "&", "|", "\t", qr/\s+/);

    my $fallback_fs;

    foreach my $fs (@field_sep){
	my $n_fields = count_fields_using_fs($data_ref, $fs);
	next if ($n_fields == 0);
	if ($n_fields >= 2){
	    return $fs;
	}
	elsif ($n_fields == 1){
	    $fallback_fs = $fs;
	}
    }
    if (defined $fallback_fs){
	return $fallback_fs;
    }
    else {
	croak("Could not determine field separator\n");
    }
}


#*********************************
sub count_fields_using_fs{

    # With "good" defined as "will split the lines of the file into an equal
    # number of columns for all lines"

    my $data_ref = shift;
    my $fs = shift;

    my $parser = Text::RecordParser->new;

    # clean leading/trailing spaces (NOT whitespace) for purpose of counting cols
    my @clean_data_ref = @{$data_ref};
    s/^ *| *$//g for @clean_data_ref;

    $parser->data( \@clean_data_ref );
    $parser->field_separator( $fs );

    my $parsed_data = $parser->fetchall_arrayref;

    my $length = scalar(@{$parsed_data->[0]});

    # not sure what to do with files of single column data
    # might not have a handy way of dealing with that unless
    # explicit...
    
    foreach my $record (@$parsed_data){
	my $linelength = scalar(@$record);
	if ($linelength == $length){
	#    print "length = $length,";
	    next;
	}
	return 0;
    }
    
    return $length;

}


#*********************************
sub table_array_to_hash{
    
    my $array_ref = shift;
    my $arr_length =  scalar(@{$array_ref});
    my @columns = keys %{$array_ref->[0]};

    my %hash_conv;

    for my $i (0 .. ($arr_length-1)){
	foreach my $header (@columns){
	    if (not defined ($hash_conv{$header})){
		$hash_conv{$header} = [];
	    }
	    push @{$hash_conv{$header}}, $array_ref->[$i]->{$header};
	}
    }

    return \%hash_conv;
}

#*********************************
sub table_hash_to_array{
    
    my $hash_ref = shift;
    my @columns = keys %{$hash_ref};
    my @array;	
    my $hash_arr_length = scalar(@{$hash_ref->{$columns[0]}});
    
    for my $i (0 .. ($hash_arr_length-1)){
	my %line;
	foreach my $header (@columns){
	    $line{$header} = $hash_ref->{$header}[$i];
	}
	push @array, \%line;
    }

    return \@array;

}

#*********************************
sub table_print_hash{

    my $hash_ref = shift;
    my @columns = keys %{$hash_ref};

    foreach my $header (@columns){
	print "$header: ";
	my $line = join(':,:', @{$hash_ref->{$header}});
	print "$line \n";
    }
}

#*********************************
sub table_print_array{

    my $array_ref = shift;
    my $arr_length = scalar(@{$array_ref});
    
    for my $i (0 .. ($arr_length-1)){
	print "line $i: ";
	while (my ($header, $data) = each %{$array_ref->[$i]}){
	    print "$header = :$data:, ";
	}
	print "\n";
    }	

}

#*********************************
sub check_rdb{

    my $data_ref = shift;
    my $opt_ref = shift;
    
    my $secondline = $data_ref->[1];
    my $is_rdb;

    if ($secondline =~ /^(\d*\w\t)+/){
	$is_rdb = 1;
    }
    if ($secondline =~ /^(-+\t)+/){
	$is_rdb = 1;
    }

    if (($opt_ref->{format} eq 'rdb') and (not $is_rdb)){
	croak "RDB specified, but does not match second line pattern \n";
    }

    if ($is_rdb){
	$opt_ref->{format} = 'rdb';
	splice(@{$data_ref}, 1, 1);
    }
    
    return ($data_ref, $opt_ref);
}

#*********************************
sub strip_data_comments{

    my $data_lines_ref = shift;
    my $opt_ref = shift;

    my $comment = $opt_ref->{comment};
    my @cleandata;
    
    # if a comment symbol has been specified
    # remove lines that start with the symbol
    # and return a reference to a new line array

    if (defined $comment){
	my @cleandata;
	@cleandata = grep(!/^$comment/, @{$data_lines_ref});
	return \@cleandata;
    }
    # if no comment symbol is specified, just
    # return the original array reference
 
    return $data_lines_ref;
}


#*********************************
sub convert_pdl_hash{

    
    my $pdl_hash = shift;
 
    my %not_pdl_hash;

    my $pdl_type;   
    
    eval 'use PDL qw/ list / ' ;
    if ($@){
	croak("!$@");
    }


    while ( my ($header, $value) = (each %{$pdl_hash})){

	if (UNIVERSAL::isa($value,'PDL')) {
	    my $pdl_dim = $value->getndims;
	    croak("Fits table contains $pdl_dim-D objects!\n") if ($pdl_dim > 1);
	    $not_pdl_hash{$header}  = [ PDL::list($pdl_hash->{$header}) ];
	}
	else{
	    # if rdfits returns a string array, make it a string
	    # and return it

	    if (ref($value) eq 'ARRAY'){
		my $pdl_string = join( qq{}, @{$value} );
		$not_pdl_hash{$header} = $pdl_string;
	    }

	    else{
		$not_pdl_hash{$header} = [ $pdl_hash->{$header} ];
	    }

	}       
    }


    return \%not_pdl_hash;

}

#*********************************
sub check_options{

    my $opt_ref = shift;

    if (defined $opt_ref->{format}){
	croak( __PACKAGE__ ,": Invalid format specified: should be 'html','fits','csv', or 'rdb' \n")
	    unless ($opt_ref->{format} =~ /(html|fits|csv|rdb)/);
    }
    
    if (defined $opt_ref->{comment} and (not defined $opt_ref->{field_separator})){
	croak( __PACKAGE__ ,": Comment symbol cannot be possible field separator \n") 
	    if ($opt_ref->{comment} =~  /(,|&|\||\\t|\s+)/ );
    }
    
    if (defined $opt_ref->{comment} and defined $opt_ref->{field_separator}){
	croak( __PACKAGE__ ,": Comment symbol cannot match field separator \n")
	    if ($opt_ref->{comment} eq $opt_ref->{field_separator});
    }
    

}

##*********************************
sub dbg  {

    if ($Debug) {
	my $args = join('',@_) || "";
	my $caller = (caller(1))[0];
	my $line = (caller(0))[2];
	$caller ||= $0;
	if (length $caller > 22) {
	    $caller = substr($caller,0,10)."..".substr($caller,-10,10);
	}
	my $align = ' 'x39;
	$args =~ s/\n/\n$align/g;
    print STDERR sprintf ("%02d:%02d:%02d [%22.22s %4.4s]  %s\n",
                          (localtime)[2,1,0],$caller,$line,$args);
    }
}


1;
__END__


=head1 NAME

Data::ParseTable - Perl extension for table parsing

=head1 SYNOPSIS

Common Usage

  use Data::ParseTable;

  $return_array_of_hashes = Data::ParseTable::parse_table( $filename );

  $return_hash_array =      Data::ParseTable::parse_table( $filename , 
                                                          { return_as_hash => 1 });

  $return_array_of_hashes = Data::ParseTable::parse_table( \@data_lines );

  $return_hash_array =      Data::ParseTable::parse_table( \@data_lines , 
                                                          { return_as_hash => 1 });



=head1 DESCRIPTION

Data::ParseTable parses text, ascii or binary fits, or html tables into arrays of hashes or hashes 
of arrays.  

=head2 EXPORT

None by default.
parse_table() is available if requested.

=head2 METHODS

Parse table is the only accessible method.  It is designed to accept either a filename or a
reference to an array of lines/strings as its first argument.  An optional second argument
may contain options such as the format of the data, its field separator, a request that its
return type of the method be a hash of arrays, and its comment symbol.

parse_table( $filename or \@data_lines,
           { format          => undef,  # File format (supported vals = fits|rdb|html|csv)
             field_separator => undef,  # Field separator for text file formats
             return_as_hash  => undef,  # Override default: return hash of arrays.
             comment         => undef,  # Comment specifier (e.g. qr/^#/)
	     debug           => 0,      # print debug info if 1
           } 
);

In the absence of options specifying format and field separator, these
will be guessed by the method.  To guess the field separator, a
subroutine steps through these field separator choices in this order:
(",", "&", "|", "\t", qr/\s+/) .  If a separator separates all the
text lines into the same number of fields and that number is greater
than one, that separator is picked.  If no separator cleanly splits
into more than 1 column, the last separator that returns 1 column is
returned.  If no separator cleanly splits all of the lines to the same
number of columns, the routine will croak.


Absence of a comment specifier when used with a file with comments
will probably result in failure of the method.  (We can't guess
comment specifiers).

=head1 AUTHOR

Jean Connelly, E<lt>jeanconn@cfa.harvard.eduE<gt>

=head1 COPYRIGHT AND LICENSE

Copyright (C) 2005 by Jean Connelly

This library is free software; you can redistribute it and/or modify
it under the same terms as Perl itself, either Perl version 5.8.0 or,
at your option, any later version of Perl 5 you may have available.


=cut
