#!/usr/bin/env perl

# Install modules for yaxx

use warnings;
use strict;
use Cwd;
use Getopt::Long;
use File::Basename;

my %opt = ( force => 0,
	 dryrun => 0,
       );

GetOptions(\%opt,
	   "dryrun!",
	   "force!",
	   "local!",
	   );

my $INSTALL = cwd;
my $BUILD   = "$INSTALL/build";
my $PKGS     = "$INSTALL/pkgs";
my $PERL5LIB = "$INSTALL/yaxx-perl";
my $CIAO_VERSION = '3.2.1'; 

$ENV{YAXX_DIR} = $INSTALL;
$ENV{PERL5LIB} = "$PERL5LIB:$PERL5LIB/lib" . ( $ENV{PERL5LIB} ? ":$ENV{PERL5LIB}" : '');
$ENV{PATH} = "$INSTALL/bin" . ( $ENV{PATH} ? ":$ENV{PATH}" : '');
$ENV{CFITSIO} = $INSTALL;

# Check for ciao (existence and version) and setenv PGPLOT_DIR
die "Error - Looks like you are not in the CIAO environment\n" unless $ENV{ASCDS_INSTALL};

open VERSION, "$ENV{ASCDS_INSTALL}/VERSION" or die "Error - Could not find CIAO VERSION file\n";
$_ = <VERSION>;
close VERSION;
if (/CIAO \s+ ([\d\.]+)/x) {
    die "Error - CIAO version of $CIAO_VERSION is required.  Found CIAO $1\n"
      unless (vers_num($1) >= vers_num($CIAO_VERSION));
} else {
    die "Error - $ENV{ASCDS_INSTALL}/VERSION file does not have expected version information\n";
}

$ENV{PGPLOT_DIR} = "$ENV{ASCDS_INSTALL}/ots/pgplot";

# Create the build directory
run("mkdir $BUILD") unless (-d $BUILD);

# Set the list of packages.  Those with two words are interpreted as
# code programs that get built with configure.  The first word is the
# tarball root name, the second is the directory that tarfile creates.
#
# The --NOTEST packages involve ds9, and seem to fail in some cases where
# the package is just fine.  So just take a chance and assume it works..

my @packages = @ARGV ? @ARGV :
                  ( 'cfitsio2510 cfitsio',
		    qw( 
		       Spiffy
		       IO::All
		       Compress::Zlib
		       Archive::Zip
		       Module::ScanDeps
		       Digest::SHA
		       List::MoreUtils
		       File::Slurp 
		       IO::Tty
		       YAML
		       Tie::IxHash
		       File::chdir
		       Clone
		       Class::MakeMethods
		       Config::General
		       Hash::Merge
		       File::Copy::Link
		       Expect      
		       Expect::Simple 
		       ExtUtils::F77
		       PGPLOT
		       PDL    	  
		       RDB         
		       Params::Validate
		       Astro::FITS::CFITSIO
		       Astro::FITS::CFITSIO::CheckStatus
		       Astro::FITS::Header
		       CFITSIO::Simple
		       Astro::FITS::CFITSIO::Simple
		       Env::Path
		       Time-modules
		       Time::DayOfYear
		       Ska::Convert
		       Ska::IO
		       Ska::Process
		       Ska::RDB
		       Ska::HashTable--NOTEST
		       Decision::Depends
		       Chandra::Tools::dmcoords
		       Module::Build
		       version
		       Readonly
		       Sub::Uplevel
		       Test::Simple
		       Test::Exception
		       IO::stringy
		       Text::RecordParser
		       Text::Glob
		       HTML::Tagset
		       HTML::Parser
		       HTML::TableExtract
		       File::MMagic
		       HTML::Table
		       Regexp::Common
		       Data::ParseTable
		       Chandra::Tools::ECF
		      ));

# Modules that require use Build.PL instead of Makefile.PL
my %use_Build_PL = map { $_ => 1 } qw(Module::Build
				      Text::RecordParser
				      version
				     );

# Try to 'use' these modules for specified packages.  The default is the package name
my %test_modules_for =  ('Time-modules' => [ qw(Time::JulianDay Time::ParseDate Time::CTime) ],
			 'IO::stringy'   => [ qw(IO::Stringy IO::AtomicFile IO::Scalar IO::Lines) ],
			);

foreach (@packages) {
    my @comp = split;
    (@comp == 1) ? perl_module_install(@comp) : code_install(@comp);
}

print "\nCreating customized executable 'yaxx' in $INSTALL\n";
chdir $INSTALL;
$^I = '~';
@ARGV = qw(yaxx);
while (<>) {
    if (/\A \s* YAXX_DIR \s* = \s* \S+/x) {
	print "YAXX_DIR='${INSTALL}'\n";
    } else {
	print;
    }
} continue { close ARGV if eof} ;

sub perl_module_install {
    my $module = shift;

    # if module name has '--NOTEST' at the end, then remove this and set flag so that
    # 'make test' is not run
    $module =~ s/(..NOTEST)//g;
    my $no_test = 0;

    # Get the package name 
    (my $pkg_name = $module) =~ s/::/-/g;

    print "\n**************** Build and install $module *****************\n\n";

    # First see if the module can be used
    # 
    print "Testing for module accessibility\n";
    my @test_modules = defined $test_modules_for{$module} ? @{$test_modules_for{$module}} : ($module);
    my $use_module_failed;
    foreach my $test_module (@test_modules) {
	print "Trying to use module $test_module ... ";
	$use_module_failed = system(qq(env PERL5LIB=$ENV{PERL5LIB} perl -e 'use $test_module' 2>&1 > /dev/null));
	print $use_module_failed ? "NOT OK\n" : "OK\n";
	last if $use_module_failed;
    }

    unless ($opt{local} or $opt{force}) {
	unless ($use_module_failed) {
	    print "  Already installed\n";
	    return;
	}
    }

    # Look for gunzipped tar files matching module name
    run_cd("$PKGS");
    my @files = glob("*.tar.gz");
    my @pkgs = grep /\A${pkg_name}-[^-]+\Z/, @files;
    die "Looking for the module ${pkg_name}-*.tar.gz in $PKGS, found '@pkgs' instead\n"
      unless (@pkgs == 1);
    (my $build = $pkgs[0]) =~ s/(_patch)?\.tar\.gz//;
	
    if (-e "$BUILD/$build/.installed") {
	if ($opt{force} or $use_module_failed) {
	    run("rm $BUILD/$build/.installed");
	} else {
	    print "  Already installed\n";
	    return;
	}
    }

    my $Makefile_PL = $use_Build_PL{$module} ? "Build.PL --install_base $PERL5LIB"
      : "Makefile.PL PREFIX=${PERL5LIB} LIB=${PERL5LIB}";
    my $make = $use_Build_PL{$module} ? "./Build" : "make";

    run_cd("$BUILD");
    run("gunzip --stdout $PKGS/$pkgs[0] | tar xvf -");
    run_cd("$build");
    run("perl $Makefile_PL");
    run("$make");
    run("$make test") unless $no_test;
    run("$make install");
    run("touch .installed");
}
    

sub code_install {
    my ($pkg, $build) = @_;
    print "\n**************** Build and install $build *****************\n\n";

    if (-e "$BUILD/$build/.installed") {
	if ($opt{force}) {
	    run("rm $BUILD/$build/.installed");
	} else {
	    print "  Already installed\n";
	    return;
	}
    }

    run_cd("$BUILD");
    run("gunzip --stdout $PKGS/${pkg}.tar.gz | tar xvf -");
    run_cd("$build");
    run("./configure --prefix=$INSTALL");
    run("make");
    run("make install");
    run("rm *.o");
    run("touch .installed");
}

sub run {
    my $cmd = shift;
    print "$cmd\n";
    return if $opt{dryrun};

    if (system($cmd)) {
	print "\n\n***** WARNING - '$cmd' appears to have failed ***\n\n";
	print " Options: (c)ontinue anyway  (q)uit : ";
	my $input = <STDIN>;
	die unless $input =~ /\A \s* c/ix;
    }
}

sub run_cd {
    my $dir = shift;
    my $cmd = "cd $dir";
    print "$cmd\n";
    return if $opt{dryrun};

    unless (chdir $dir) {
	print "\n\n***** ERROR - '$cmd' did not succeed ***\n\n";
	die;
    }
}

sub vers_num {
# Convert a string version to a numerical version, e.g.
# 3.2.53 -> 3.002053
    my $v = shift;
    my @v = split /[\.]/, $v;
    my $mult = 1;
    my $version = 0;
    local $_;
    foreach (@v) {
	$version += $_ * $mult;
	$mult /= 1000;
    }
    return $version;
}
