# --8<--8<--8<--8<--
#
# Copyright (C) 2006 Smithsonian Astrophysical Observatory
#
# This file is part of yaxx
#
# yaxx is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# yaxx is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the 
#       Free Software Foundation, Inc. 
#       51 Franklin Street, Fifth Floor
#       Boston, MA  02110-1301, USA
#
# -->8-->8-->8-->8--

package YaxxSource;

use strict;
use warnings;

# 'Local' CXC packages
use Ska::Process qw(message run_tool make_local_copy);
use Ska::Convert qw(dec2hms);
use CFITSIO::Simple;
use Chandra::Tools::dmcoords;
use Chandra::Tools::ECF;

# Available on CPAN
use HTML::Entities ();
use HTML::Table;
use IO::All;
use Regexp::Common;
use Decision::Depends;
use Data::Dumper;
use Data::ParseTable;
use POSIX qw(tmpnam strtod);
use Carp;
use File::Basename;
use File::Spec;
use File::Slurp;
use File::Copy;
use File::Glob qw(:glob);
use File::chdir;
use Cwd;
use English;
use Storable qw(dclone);
use Config::General;
use Hash::Merge qw(merge);
use PDL;
use PDL::Graphics::LUT;
use PDL::Graphics::PGPLOT::Window;
use PGPLOT;
use List::MoreUtils;
use Text::Glob;

use Class::MakeMethods::Standard::Hash ( 
  scalar => [ qw( 
		 allow_failed_fit 
		 asol_glob 
		 analysis_dir 
		 badpix_glob
		 bgd_ann_sep 
		 bgd_ann_wid 
		 binned_method 
		 binned_stat
		 common_model_defs
		 config_file counts
		 evt2_glob exposure 
		 fakeid
		 file_definition 
		 fit_plot_size
		 fit_rules
		 force
		 group_val
		 input_dir 
		 lock 
		 max_energy 
		 min_counts 
		 min_energy 
		 min_src_rad model 
		 formatting_rule
		 next_source
		 no_source_image 
		 obsid 
		 preclean
		 prev_source
		 process 
		 process_fake
		 program_dir
		 projection
		 psf_fraction 
		 psf_energy
		 report_index_col
		 sherpa_cmds_grouped 
		 sherpa_cmds_ungrouped 
		 source_data 
		 source_image 
		 spec_chan_type 
		 srcid
		 src2_glob
		 src_reg 
		 summary_header_key_col
		 timeout 
		 unbinned_method 
		 unbinned_stat 
		 uncertainty
		 use_psextract
		 write_MDL_header_keys
		 fake_program						       
		 loud
		 log_time_format
		 output_dir
		 objlist						       
		 bkglist 
		 log_file
		 src_column_name
		) ]
				       );

our $FATAL_ERROR = <<'MESSAGE';

A fatal error has occurred.  If you believe this is not due to
faulty user inputs, please copy error message and the stack trace 
below and send to

  taldcroft@cfa.harvard.edu

MESSAGE

# our %formatting_rule;
our $NAN = (-1)**0.5;		# Generate a NaN 
our $DMcoord;
our $DMcoord_obsid = -1;
our %Bkg;			# Hash {obsid}{srcid} of bkg YaxxSource objects
our $ECF;			# Chandra::Tools::ECF (Encircled Counts Fraction) object
our $Arcsec_per_pixel = 0.492;	# Arcsec per ACIS pixel

Hash::Merge::set_behavior( 'RIGHT_PRECEDENT' );

# Include package code from these files
for (qw(accessors dmcoord extract fakes fit file_util group_pi image regions report)) {
    require "$_.pl";
}

##****************************************************************************
sub add_bkg_source {
# This package method will add a background source to the package global %bkg
# for use later in defining bkg region    
##****************************************************************************
    my $source = shift;
    my $obsid = $source->{obsid};
    my $srcid = $source->{srcid};
    if (not defined $Bkg{$obsid}{$srcid}) {
	$Bkg{$obsid}{$srcid} = $source; 
    }
}

##****************************************************************************
sub new {
##****************************************************************************
    my $classname = shift;
    my $self = {};
    bless ($self, $classname);
    local $_;

    # Merge the default config, project config, and source info hashes into $par
    # using Hash::Merge routine with right precedence
    my $par = {};
    map { $par = merge($par, $_) } @_;

    # Set object data required for $self->file() to find other yaxx.cfg files
    $self->$_( $par->{$_} ) foreach qw (file_definition obsid srcid fakeid analysis_dir config_file);

    # Merge config (yaxx.cfg) from obsid and source directory
    for ($self->file('obsid_dir') . '/' . $self->config_file,
	 $self->file('src_dir') . '/' . $self->config_file ) {
	$par = merge( $par, { ParseConfig(-ConfigFile => $_) } ) if (-r $_);
    }

    # Make sure spec chan type is lowercase 
    $_ = lc for $par->{spec_chan_type}; 

    # Finally set all object data via accessor methods
    foreach (keys %{$par}) {
	eval { $self->$_( $par->{$_} ) };
	if ($@) {
	    message("ERROR - user supplied config parameter '$_' is not allowed\n");
	    confess $YaxxSource::FATAL_ERROR;		# So long
	}
    }

    return $self;
}

1;

