# --8<--8<--8<--8<--
#
# Copyright (C) 2006 Smithsonian Astrophysical Observatory
#
# This file is part of yaxx
#
# yaxx is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# yaxx is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the 
#       Free Software Foundation, Inc. 
#       51 Franklin Street, Fifth Floor
#       Boston, MA  02110-1301, USA
#
# -->8-->8-->8-->8--

use strict;
use warnings;

##****************************************************************************
sub make_dirs {
##****************************************************************************
    my $self = shift;

    my $src_dir = $self->file('src_dir');
    my $status = system("mkdir -p $src_dir"); # Create directories if needed.

    if ($status) {
	message("ERROR - Could not create directory $src_dir\n");
	confess $YaxxSource::FATAL_ERROR;
    }

    1;
}

##****************************************************************************
sub clean {
##****************************************************************************
    my $self = shift;
    my $type_list = shift;
    local $_;
    return 1 unless $type_list;

    my $src_dir = $self->file('src_dir');
    chdir $src_dir;

    # Find all the files associated with spectral fitting
    my @fit_files;
    foreach my $mdl (map {basename $_,'.in'} glob("*.in")) {
	push @fit_files, map {"$mdl.$_"} qw(mdl in ps unc proj jpg);
    }

    # Define how to expand the pseudo-types, either in terms of base types
    # or by a list of files
    my %clean_types = ( all => [qw(obsid_evt2 evt2 obsid_asol asol
				   extract fit source_image log report resources_dir)
			       ],
			extract => [qw(arf rmf asphist pi pi_bin bg_pi bg_arf bg_rmf)],
		      );
    my %clean_files = ( report       => [ glob($self->file('report').".*") ],
			source_image => [ glob($self->file('source_image').".*") ],
			log    	     => [ glob($self->file('log').".*") ],
			region 	     => [ glob("*.reg") ],
			resource_dir => [ glob($self->file('resources_dir') . "/*") ],
			fit    	     => [ @fit_files ],
		      );

    my @clean;
    my @type = split ' ', $type_list;
    while (my $type = shift @type) {
	# Try to:
	#   1) expand the type into component types
	#   2) find a list of files for that type
	#   3) find the single file for that type
	if ($clean_types{$type}) {
	    push @type, @{$clean_types{$type}};
	} elsif ($clean_files{$type}) {
	    push @clean, @{$clean_files{$type}} if @{$clean_files{$type}};
	} else {
	    my $file;
	    eval { $file = $self->file($type) };
	    if ($@) {		
		message("WARNING - Unknown file type '$type' for cleaning\n");
	    } else {
		push @clean, $file if defined $file;
	    }
	}
    }

    if (@clean) {
	message("Cleaning files:\n   " . join("\n   ",@clean) . "\n");
	map { unlink $_ if (-e $_ or -l $_) } @clean;
    }

    1;
}

##****************************************************************************
sub get_lock {
##****************************************************************************
    my $self = shift;
    my $msg;

    return 1 unless $self->lock;

    my $lock_file = $self->file('lock');
    my $date = gmtime();

    if (-r $lock_file) {
	$msg = read_file($lock_file);
	message($msg);
	return;
    } else {
	$msg = "Source locked by $ENV{USER} at $date by process $PROCESS_ID on $ENV{HOST}\n";
	write_file($lock_file, $msg);
    }

    return 1;
}

##****************************************************************************
sub release_lock {
##****************************************************************************
    my $self = shift;

    my $lock_file = $self->file('lock');
    return ((-w $lock_file and unlink $lock_file) or not $self->lock);
}

##****************************************************************************
sub file {
#
# Encapsulate all the naming conventions here, using the file_definition
# configuration variable
#
##****************************************************************************
    my ($self, $arg, $opt) = @_;
    my $cwd = $opt->{cwd} || cwd;

    my ($type, @ext) = split /\./, $arg;

    my $f = $self->file_definition->{$type};
    if (not defined $f) {
	message "ERROR - invalid file type '$type' detected\n";
	confess $YaxxSource::FATAL_ERROR;
    }

    # Strip any whitespace which is inserted to improve readability
    $f =~ s/\s//g;
    $f = $self->interpolate_string($f);
    $f .= join('.', '', @ext) if @ext;

    # Return the full file path if absolute, else prepend the analysis_dir (user's
    # analysis root directory) and return relative path
    my $filepath = io($f)->is_absolute ? $f : io->join($self->analysis_dir, $f)->abs2rel;
    $filepath = '.' unless $filepath =~ /\S/;

    return $filepath;
}

########################################################################################
sub get_resource_files {
# Make sure that all yaxx resource files are available in the analysis directory
########################################################################################
    my $self = shift;

    my $resources_dir = $self->file('resources_dir');
    io($resources_dir)->mkpath unless (-d $resources_dir);

    for my $yaxx_file (io($self->file('resources_yaxx_dir'))->all ) {
	my $local_file =  io("$resources_dir/" . $yaxx_file->filename);
	# Copy unless local file exists and local file is newer than yaxx file
	$yaxx_file > $local_file unless ($local_file->exists
					 and $local_file->mtime > $yaxx_file->mtime );
    }

    return 1;
}

########################################################################################
sub get_data_files {
########################################################################################
    my $self = shift;
    my $cwd = cwd;
    local $_;
    my @files;

    my $identify = "obsid=".$self->obsid." srcid=".$self->srcid;
    
    # Go through globs one at a time, looking for matching files
    # Use funky bsd_glob to ensure correct behavior if the glob doesn't contain wildcards.
    # The second argument overrides the default flags that make glob return the pattern
    # if it does not contain a wildcard, even if that file doesn't exist.  Why???
    foreach (split ' ',$self->evt2_glob) {
	last if (@files = bsd_glob($self->input_dir."/$_", undef));
    }

    # Make sure we got one and only one event file
    if (@files > 1 or @files == 0) {
	my $err = @files ? 'too many' : 'no';
	message("ERROR - $err input event files '".$self->evt2_glob."' for $identify\n");
	return;
    } 

    # Make a local unzipped copy of the event file in the yaxx directory.  If the
    # event file in the database is already unzipped, then just link to it
    # Return an absolute (full) filename

    make_local_copy($files[0], $self->file('obsid_evt2')); 
    make_local_copy($self->file('obsid_evt2'), $self->file('evt2'));
    
    # Now find the aspect asol/aoff files for the obsid
    foreach (split ' ',$self->asol_glob) {
	last if (@files = bsd_glob($self->input_dir."/$_", undef));
    }

    # Find the aspect offset file and make a local unzipped copy in obsid directory
    my $obsid_asol = $self->file('obsid_asol');
    if (@files == 0) {
	message("ERROR - no input AOFF/ASOL files for $identify\n");
	return;
    } elsif (@files == 1) {
	make_local_copy($files[0], $obsid_asol); # One per obsid
    } else {
        # Run dmmerge if necessary
	# Delete existing file (to force rebuild) if it is older than any of the input files

	if (test_dep(-target => $obsid_asol, -depend => \@files)) {
	    # Multiple files, need to merge them
	    my $list = join ',', @files;

	    run_tool("dmmerge",
		     infile => $list,
		     columnList => '',
		     outfile => $self->file('obsid_asol'),
		     outBlock => '',
		     clobber => 'yes',
		     lookupTab => "$ENV{ASCDS_CALIB}/dmmerge_header_lookup.txt",
		     { timeout => $self->timeout,
		       paste => 1,
		       loud => 1,
		     }
		    ) or return;
	}
    }

    # Make a local linked copy in the source directory
    make_local_copy($self->file('obsid_asol'), $self->file('asol')); 

    # Now find the bad pixel files [optional]
    foreach (split ' ',$self->badpix_glob) {
	last if (@files = bsd_glob($self->input_dir."/$_",undef));
    }

    # Throw a warning if multiple bad pixel files found
    if (@files > 1) {
	my $err = @files ? 'too many' : 'no';
	message("ERROR - too many bad pixel files '".$self->badpix_glob."' for $identify\n");
	return;
    } elsif (@files == 1) {
	make_local_copy($files[0], $self->file('obsid_badpix')); # One per obsid
	make_local_copy($self->file('obsid_badpix'), $self->file('badpix')); # One per obsid
    }

    return (-e $self->file('asol') and -e $self->file('evt2')); # actual success?
} 

#********************************************************************************
sub store_object {
# Store a copy of the object (to a depth of 1) in the object_dump file
# in the source directory
#********************************************************************************
    my $self = shift;
    
    local $Data::Dumper::Maxdepth = 1;
    my $dumper = Dumper $self;
    eval { $dumper > io($self->file('object_dump')) };
    if ($@) {
	message("ERROR - $@\n");
	confess $YaxxSource::FATAL_ERROR;
    }
    return 1;
}

#********************************************************************************
sub restore_object {
# Restore the stored object and return it
#********************************************************************************
    my $self = shift;
    
    my $dumper_output;
    my $dumper_file = $self->file('object_dump');
    eval { $dumper_output < io($dumper_file) };
    if ($@) {
	message("WARNING - unable to find stored information file for ObsId "
		. $self->{obsid} . ", Src " . $self->{srcid}
		. ".  Was it processed?\n");
	return;
    }

    # Eval the expression in the file, which sets a variable that gets
    # returned as the output of eval
    return eval "my $dumper_output";
}    

#********************************************************************************
sub open_log {
#
# Open a log file (using 'message') for processing related to this source
#********************************************************************************
    my $self = shift;
    my $log = $self->file('log');
    local $_;

    if (-e $log) {
	my $max = 1;
	map { $max = $1+1 if (/${log}\.(\d+)/ and $1 >= $max) } glob("${log}*");
	move($log, "$log.$max");
    }
    message('', (init  => 1,
		 log_file => $log));
    1;
}

#********************************************************************************
sub close_log {
#********************************************************************************
    my $self = shift;
    message('', (close => 1, 
		 log_file => $self->file('log')));
}


1;
