# --8<--8<--8<--8<--
#
# Copyright (C) 2006 Smithsonian Astrophysical Observatory
#
# This file is part of yaxx
#
# yaxx is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# yaxx is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the 
#       Free Software Foundation, Inc. 
#       51 Franklin Street, Fifth Floor
#       Boston, MA  02110-1301, USA
#
# -->8-->8-->8-->8--

use strict;
use warnings;

##****************************************************************************
sub make_images {
# Use PGPLOT to make color image map of source and immediate surrounding
# and the source extraction / background regions
##****************************************************************************
    my $self = shift;
    my $img_opt = $self->source_image;
    local $_;

    # Skip images if this is a fake dataset
    return 1 if ($self->fakeid or $self->no_source_image);

    # Define various file names for use later
    my $image_file = $self->file('source_image.ps');
    my $evt = File::Spec->rel2abs($self->file('evt2'));
    my $img_tmp = tmpnam();

    my $src = $self->file('src');
    my $bkg = $self->file('bkg');
    my @target = ($image_file, $self->file('source_image.jpg'));
    my @test_dep = (-target => \@target, -depend => [$evt, $src, $bkg]);

    # Check on dependencies 
    return 1 unless (test_dep(@test_dep));
    map {unlink if -e} (@target);  # Clean target files

    my $x = sprintf("%d", $self->X);
    my $y = sprintf("%d", $self->Y);
    my @reg;

    my $sz2 = $img_opt->{field_of_view} / 2;
    my $energy = "energy=".$self->min_energy.":".$self->max_energy;
    run_tool(sprintf("dmcopy '%s[bin x=%d:%d,y=%d:%d][$energy]' $img_tmp",
		     $evt, $x-$sz2, $x+$sz2, $y-$sz2, $y+$sz2), {loud=>0});
    my $win = PDL::Graphics::PGPLOT::Window->new(device => "$image_file/vcps",
						 size => $img_opt->{size_inches},
						 unit => 'inch', # Inches
						 hold => 1,
						 axis =>  'box',  # Make plain (empty) axes
						 # ['',''], # ?

						);
    $win->ctab(lut_data($img_opt->{color_table}, $img_opt->{invert}));
    my $img = rfits($img_tmp);
    
    # Force fits_imag to use Physical (sky) WCS coords instead of (RA,Dec)
    # which is the default
    for my $axis (1, 2) {
	for my $hdrkey (qw(CTYPE CRVAL CRPIX CDELT)) {
	    $img->hdr->{"${hdrkey}${axis}"} = $img->hdr->{"${hdrkey}${axis}P"};
	}
    }
    $win->fits_imag($img, {# WCS => 'P',   # This used to work (perhaps a PDL version issue?)
			                   # but has been replaced by above code
			   itf => 'log',
			   DrawWedge => 0,
			   xtitle => ' ',
			   ytitle=> ' ',
			  });
    $win->hold(1);

    # Add text label for object name.  Could not figure out how to get limits from Window
    # interface, so use PGPLOT pgqwin routine directly
    if ($img_opt->{label}) {
	my @lims = (undef,undef,undef,undef);
	my $label = $self->interpolate_string($img_opt->{label});
	pgqwin(@lims);
	my $text_x = $lims[0] + ($lims[1]-$lims[0]) * $img_opt->{label_x};
	my $text_y = $lims[2] + ($lims[3]-$lims[2]) * $img_opt->{label_y};
	$win->text($label, $text_x, $text_y, 
		   {charsize => $img_opt->{charsize},
		    justification => $img_opt->{justification}
		   }
		  );
    }

    foreach (ciao_reg_to_pgplot($src), ciao_reg_to_pgplot($bkg)) {
	my $pg_sub = $_->{pg_sub};
	my $pg_arg = $_->{pg_arg};
	$win->$pg_sub(@{$pg_arg}, { filltype => 'outline' } );
    }

    $win->close();
    unlink $img_tmp;

    # Convert to jpeg by using the eps2png perl script
    my $width = $img_opt->{size_pixels};
    my $image_file_jpeg = $self->file('source_image.jpg');
    run_tool("eps2png -jpeg -output $image_file_jpeg -width $width $image_file",
	    {loud=>0});

    if (test_dep(@test_dep)) {
	message("ERROR - failed to make source image\n");
	return;
    }

    1;
}

##****************************************************************************
sub ciao_reg_to_pgplot {
# Convert CIAO region file to PGPLOT commands
##****************************************************************************
    my $file = shift;
    my @pg;			# PGPLOT subroutine/argument specifiers
    my $deg2rad = sclr atan(1) / 45;
    local $_;
#                          ----------------------------------------------------------
#                             |                     Region Shapes                      |
#                             |--------------------------------------------------------|
#                             | RECTangle(xmin,ymin,xmax,ymax)                         |
#                             |--------------------------------------------------------|
#                             | BOX(xcenter,ycenter,xlen,ylen)                         |
#                             |--------------------------------------------------------|
#                             | ROTBOX(xcenter,ycenter,xlen,ylen,angle)                |
#                             |--------------------------------------------------------|
#                             | POLYgon(x1,y1,x2,y2,x3,y3,...)                         |
#                             |--------------------------------------------------------|
#                             | CIRcle(xcenter,ycenter,radius)                         |
#                             |--------------------------------------------------------|
#                             | ANNULUS(xcenter,ycenter,iradius,oradius)               |
#                             |--------------------------------------------------------|
#                             | ELLipse(xcenter,ycenter,xradius,yradius,angle)         |
#                             |--------------------------------------------------------|
#                             | SECTor(xcenter,ycenter,minangle,maxangle)              |
#                             |--------------------------------------------------------|
#                             | PIE(xcenter,ycenter,iradius,oradius,minangle,maxangle) |
#                             |--------------------------------------------------------|
#                             | POInt(xcenter,ycenter)                                 |
#                             |--------------------------------------------------------|
#                             | FIELD()                                                |
#                             |--------------------------------------------------------|
#                             | REGION(file)                                           |
#                             ----------------------------------------------------------

    # Parse the region files and build the PGPLOT commands for all regions
    $_ = read_file($file);
    s/#[^\n\r]*//g;           # Strip out any comments
    while (m{ [-+]? \s* (\w+) \( ([^()]+) \) }xg) {
	my ($pg_sub, $pg_arg);
	my ($shape, $arg_string) = ($1, $2);
	my @arg = split /\s*,\s*/, $arg_string;
	if ($shape =~ /^rect/i) {
	    $pg_sub = 'rectangle';
	    $pg_arg = [ ($arg[2]+$arg[0])/2,  # xcenter
			($arg[3]+$arg[1])/2,  # ycenter
			($arg[2]-$arg[0]),    # xside
			($arg[3]-$arg[1])     # yside
		      ];
	} elsif ($shape =~ /^box/i) {
	    $pg_sub = 'rectangle';
	    $pg_arg = [ @arg[0..3] ];
	} elsif ($shape =~ /^rotbox/i) {
	    $pg_sub = 'rectangle';
	    $pg_arg = [ @arg[0..3], $arg[4]*$deg2rad ];
	} elsif ($shape =~ /^poly/i) {
	    $pg_sub = 'poly';
	    my $n = POSIX::floor(@arg / 2);
	    my @x = map { $arg[$_*2] } (0..$n-1);
	    my @y = map { $arg[$_*2+1] } (0..$n-1);
	    $pg_arg = [ \@x, \@y ];
	} elsif ($shape =~ /^cir/i) {
	    $pg_sub = 'circle';
	    $pg_arg = [ @arg[0..2] ];
	} elsif ($shape =~ /^annulus/) {
	    $pg_sub = 'circle';
	    $pg_arg = [ @arg[0..1], $arg[2] ];
	    push @pg, {pg_sub => $pg_sub,
		       pg_arg => $pg_arg,
		      };
	    $pg_arg = [ @arg[0..1], $arg[3] ];
	} elsif ($shape =~ /^ell/) {
	    $pg_sub = 'ellipse';
	    $pg_arg = [ @arg[0..3], $arg[4]*$deg2rad ];
	} else {
	    message("Region $shape\($arg_string\) not supported - ignoring\n");
	    next;
	}
	push @pg, {pg_sub => $pg_sub,
		   pg_arg => $pg_arg,
		  };

    }

    return @pg;
}

1;
